@@ -3,4 +3,4 @@
 #endif
 #include "system.h"
 #include "grep.h"
-char const default_matcher[] = "egrep";
+char const *matcher = "egrep";
